@@ -91,6 +91,11 @@ class CodeMetadataExprQueue {
 };
 
 class BinaryReaderIR : public BinaryReaderNop {
+  static constexpr size_t kMaxNestingDepth = 16384;  // max depth of label stack
+  static constexpr size_t kMaxFunctionLocals = 50000;  // matches V8
+  static constexpr size_t kMaxFunctionParams = 1000;   // matches V8
+  static constexpr size_t kMaxFunctionResults = 1000;  // matches V8
+
  public:
   BinaryReaderIR(Module* out_module, const char* filename, Errors* errors);
 
@@ -350,9 +355,9 @@ class BinaryReaderIR : public BinaryReaderNop {
  private:
   Location GetLocation() const;
   void PrintError(const char* format, ...);
-  void PushLabel(LabelType label_type,
-                 ExprList* first,
-                 Expr* context = nullptr);
+  Result PushLabel(LabelType label_type,
+                   ExprList* first,
+                   Expr* context = nullptr);
   Result BeginInitExpr(ExprList* init_expr);
   Result EndInitExpr();
   Result PopLabel();
@@ -404,10 +409,15 @@ void WABT_PRINTF_FORMAT(2, 3) BinaryReaderIR::PrintError(const char* format,
   errors_->emplace_back(ErrorLevel::Error, Location(kInvalidOffset), buffer);
 }
 
-void BinaryReaderIR::PushLabel(LabelType label_type,
-                               ExprList* first,
-                               Expr* context) {
+Result BinaryReaderIR::PushLabel(LabelType label_type,
+                                 ExprList* first,
+                                 Expr* context) {
+  if (label_stack_.size() >= kMaxNestingDepth) {
+    PrintError("label stack exceeds max nesting depth");
+    return Result::Error;
+  }
   label_stack_.emplace_back(label_type, first, context);
+  return Result::Ok;
 }
 
 Result BinaryReaderIR::PopLabel() {
@@ -439,6 +449,10 @@ Result BinaryReaderIR::TopLabelExpr(LabelNode** label, Expr** expr) {
   CHECK_RESULT(TopLabel(label));
   LabelNode* parent_label;
   CHECK_RESULT(GetLabelAt(&parent_label, 1));
+  if (parent_label->exprs->empty()) {
+    PrintError("TopLabelExpr: parent label has empty expr list");
+    return Result::Error;
+  }
   *expr = &parent_label->exprs->back();
   return Result::Ok;
 }
@@ -498,10 +512,29 @@ Result BinaryReaderIR::OnFuncType(Index index,
                                   Type* param_types,
                                   Index result_count,
                                   Type* result_types) {
-  auto field = MakeUnique<TypeModuleField>(GetLocation());
-  auto func_type = MakeUnique<FuncType>();
+  if (param_count > kMaxFunctionParams) {
+    PrintError("FuncType param count exceeds maximum value");
+    return Result::Error;
+  }
+
+  if (result_count > kMaxFunctionResults) {
+    PrintError("FuncType result count exceeds maximum value");
+    return Result::Error;
+  }
+
+  auto field = std::make_unique<TypeModuleField>(GetLocation());
+  auto func_type = std::make_unique<FuncType>();
   func_type->sig.param_types.assign(param_types, param_types + param_count);
   func_type->sig.result_types.assign(result_types, result_types + result_count);
+
+  module_->features_used.simd |=
+      std::any_of(func_type->sig.param_types.begin(),
+                  func_type->sig.param_types.end(),
+                  [](auto x) { return x == Type::V128; }) ||
+      std::any_of(func_type->sig.result_types.begin(),
+                  func_type->sig.result_types.end(),
+                  [](auto x) { return x == Type::V128; });
+
   field->type = std::move(func_type);
   module_->AppendField(std::move(field));
   return Result::Ok;
@@ -510,12 +543,13 @@ Result BinaryReaderIR::OnFuncType(Index index,
 Result BinaryReaderIR::OnStructType(Index index,
                                     Index field_count,
                                     TypeMut* fields) {
-  auto field = MakeUnique<TypeModuleField>(GetLocation());
-  auto struct_type = MakeUnique<StructType>();
+  auto field = std::make_unique<TypeModuleField>(GetLocation());
+  auto struct_type = std::make_unique<StructType>();
   struct_type->fields.resize(field_count);
   for (Index i = 0; i < field_count; ++i) {
     struct_type->fields[i].type = fields[i].type;
     struct_type->fields[i].mutable_ = fields[i].mutable_;
+    module_->features_used.simd |= (fields[i].type == Type::V128);
   }
   field->type = std::move(struct_type);
   module_->AppendField(std::move(field));
@@ -523,10 +557,11 @@ Result BinaryReaderIR::OnStructType(Index index,
 }
 
 Result BinaryReaderIR::OnArrayType(Index index, TypeMut type_mut) {
-  auto field = MakeUnique<TypeModuleField>(GetLocation());
-  auto array_type = MakeUnique<ArrayType>();
+  auto field = std::make_unique<TypeModuleField>(GetLocation());
+  auto array_type = std::make_unique<ArrayType>();
   array_type->field.type = type_mut.type;
   array_type->field.mutable_ = type_mut.mutable_;
+  module_->features_used.simd |= (type_mut.type == Type::V128);
   field->type = std::move(array_type);
   module_->AppendField(std::move(field));
   return Result::Ok;
@@ -544,12 +579,12 @@ Result BinaryReaderIR::OnImportFunc(Index import_index,
                                     std::string_view field_name,
                                     Index func_index,
                                     Index sig_index) {
-  auto import = MakeUnique<FuncImport>();
+  auto import = std::make_unique<FuncImport>();
   import->module_name = module_name;
   import->field_name = field_name;
   SetFuncDeclaration(&import->func.decl, Var(sig_index, GetLocation()));
   module_->AppendField(
-      MakeUnique<ImportModuleField>(std::move(import), GetLocation()));
+      std::make_unique<ImportModuleField>(std::move(import), GetLocation()));
   return Result::Ok;
 }
 
@@ -559,13 +594,13 @@ Result BinaryReaderIR::OnImportTable(Index import_index,
                                      Index table_index,
                                      Type elem_type,
                                      const Limits* elem_limits) {
-  auto import = MakeUnique<TableImport>();
+  auto import = std::make_unique<TableImport>();
   import->module_name = module_name;
   import->field_name = field_name;
   import->table.elem_limits = *elem_limits;
   import->table.elem_type = elem_type;
   module_->AppendField(
-      MakeUnique<ImportModuleField>(std::move(import), GetLocation()));
+      std::make_unique<ImportModuleField>(std::move(import), GetLocation()));
   return Result::Ok;
 }
 
@@ -574,12 +609,12 @@ Result BinaryReaderIR::OnImportMemory(Index import_index,
                                       std::string_view field_name,
                                       Index memory_index,
                                       const Limits* page_limits) {
-  auto import = MakeUnique<MemoryImport>();
+  auto import = std::make_unique<MemoryImport>();
   import->module_name = module_name;
   import->field_name = field_name;
   import->memory.page_limits = *page_limits;
   module_->AppendField(
-      MakeUnique<ImportModuleField>(std::move(import), GetLocation()));
+      std::make_unique<ImportModuleField>(std::move(import), GetLocation()));
   return Result::Ok;
 }
 
@@ -589,13 +624,14 @@ Result BinaryReaderIR::OnImportGlobal(Index import_index,
                                       Index global_index,
                                       Type type,
                                       bool mutable_) {
-  auto import = MakeUnique<GlobalImport>();
+  auto import = std::make_unique<GlobalImport>();
   import->module_name = module_name;
   import->field_name = field_name;
   import->global.type = type;
   import->global.mutable_ = mutable_;
   module_->AppendField(
-      MakeUnique<ImportModuleField>(std::move(import), GetLocation()));
+      std::make_unique<ImportModuleField>(std::move(import), GetLocation()));
+  module_->features_used.simd |= (type == Type::V128);
   return Result::Ok;
 }
 
@@ -604,12 +640,13 @@ Result BinaryReaderIR::OnImportTag(Index import_index,
                                    std::string_view field_name,
                                    Index tag_index,
                                    Index sig_index) {
-  auto import = MakeUnique<TagImport>();
+  auto import = std::make_unique<TagImport>();
   import->module_name = module_name;
   import->field_name = field_name;
   SetFuncDeclaration(&import->tag.decl, Var(sig_index, GetLocation()));
   module_->AppendField(
-      MakeUnique<ImportModuleField>(std::move(import), GetLocation()));
+      std::make_unique<ImportModuleField>(std::move(import), GetLocation()));
+  module_->features_used.exceptions = true;
   return Result::Ok;
 }
 
@@ -621,7 +658,7 @@ Result BinaryReaderIR::OnFunctionCount(Index count) {
 }
 
 Result BinaryReaderIR::OnFunction(Index index, Index sig_index) {
-  auto field = MakeUnique<FuncModuleField>(GetLocation());
+  auto field = std::make_unique<FuncModuleField>(GetLocation());
   Func& func = field->func;
   SetFuncDeclaration(&func.decl, Var(sig_index, GetLocation()));
   module_->AppendField(std::move(field));
@@ -638,7 +675,7 @@ Result BinaryReaderIR::OnTableCount(Index count) {
 Result BinaryReaderIR::OnTable(Index index,
                                Type elem_type,
                                const Limits* elem_limits) {
-  auto field = MakeUnique<TableModuleField>(GetLocation());
+  auto field = std::make_unique<TableModuleField>(GetLocation());
   Table& table = field->table;
   table.elem_limits = *elem_limits;
   table.elem_type = elem_type;
@@ -654,7 +691,7 @@ Result BinaryReaderIR::OnMemoryCount(Index count) {
 }
 
 Result BinaryReaderIR::OnMemory(Index index, const Limits* page_limits) {
-  auto field = MakeUnique<MemoryModuleField>(GetLocation());
+  auto field = std::make_unique<MemoryModuleField>(GetLocation());
   Memory& memory = field->memory;
   memory.page_limits = *page_limits;
   module_->AppendField(std::move(field));
@@ -669,11 +706,12 @@ Result BinaryReaderIR::OnGlobalCount(Index count) {
 }
 
 Result BinaryReaderIR::BeginGlobal(Index index, Type type, bool mutable_) {
-  auto field = MakeUnique<GlobalModuleField>(GetLocation());
+  auto field = std::make_unique<GlobalModuleField>(GetLocation());
   Global& global = field->global;
   global.type = type;
   global.mutable_ = mutable_;
   module_->AppendField(std::move(field));
+  module_->features_used.simd |= (type == Type::V128);
   return Result::Ok;
 }
 
@@ -698,7 +736,7 @@ Result BinaryReaderIR::OnExport(Index index,
                                 ExternalKind kind,
                                 Index item_index,
                                 std::string_view name) {
-  auto field = MakeUnique<ExportModuleField>(GetLocation());
+  auto field = std::make_unique<ExportModuleField>(GetLocation());
   Export& export_ = field->export_;
   export_.name = name;
   export_.var = Var(item_index, GetLocation());
@@ -709,7 +747,8 @@ Result BinaryReaderIR::OnExport(Index index,
 
 Result BinaryReaderIR::OnStartFunction(Index func_index) {
   Var start(func_index, GetLocation());
-  module_->AppendField(MakeUnique<StartModuleField>(start, GetLocation()));
+  module_->AppendField(
+      std::make_unique<StartModuleField>(start, GetLocation()));
   return Result::Ok;
 }
 
@@ -727,12 +766,18 @@ Result BinaryReaderIR::OnFunctionBodyCount(Index count) {
 Result BinaryReaderIR::BeginFunctionBody(Index index, Offset size) {
   current_func_ = module_->funcs[index];
   current_func_->loc = GetLocation();
-  PushLabel(LabelType::Func, &current_func_->exprs);
-  return Result::Ok;
+  return PushLabel(LabelType::Func, &current_func_->exprs);
 }
 
 Result BinaryReaderIR::OnLocalDecl(Index decl_index, Index count, Type type) {
   current_func_->local_types.AppendDecl(type, count);
+
+  if (current_func_->GetNumLocals() > kMaxFunctionLocals) {
+    PrintError("function local count exceeds maximum value");
+    return Result::Error;
+  }
+
+  module_->features_used.simd |= (type == Type::V128);
   return Result::Ok;
 }
 
@@ -742,6 +787,7 @@ Result BinaryReaderIR::OnOpcode(Opcode opcode) {
   if (metadata) {
     return AppendExpr(std::move(metadata));
   }
+  module_->features_used.simd |= (opcode.GetResultType() == Type::V128);
   return Result::Ok;
 }
 
@@ -749,7 +795,7 @@ Result BinaryReaderIR::OnAtomicLoadExpr(Opcode opcode,
                                         Index memidx,
                                         Address alignment_log2,
                                         Address offset) {
-  return AppendExpr(MakeUnique<AtomicLoadExpr>(
+  return AppendExpr(std::make_unique<AtomicLoadExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
@@ -757,7 +803,7 @@ Result BinaryReaderIR::OnAtomicStoreExpr(Opcode opcode,
                                          Index memidx,
                                          Address alignment_log2,
                                          Address offset) {
-  return AppendExpr(MakeUnique<AtomicStoreExpr>(
+  return AppendExpr(std::make_unique<AtomicStoreExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
@@ -765,7 +811,7 @@ Result BinaryReaderIR::OnAtomicRmwExpr(Opcode opcode,
                                        Index memidx,
                                        Address alignment_log2,
                                        Address offset) {
-  return AppendExpr(MakeUnique<AtomicRmwExpr>(
+  return AppendExpr(std::make_unique<AtomicRmwExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
@@ -773,7 +819,7 @@ Result BinaryReaderIR::OnAtomicRmwCmpxchgExpr(Opcode opcode,
                                               Index memidx,
                                               Address alignment_log2,
                                               Address offset) {
-  return AppendExpr(MakeUnique<AtomicRmwCmpxchgExpr>(
+  return AppendExpr(std::make_unique<AtomicRmwCmpxchgExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
@@ -781,47 +827,46 @@ Result BinaryReaderIR::OnAtomicWaitExpr(Opcode opcode,
                                         Index memidx,
                                         Address alignment_log2,
                                         Address offset) {
-  return AppendExpr(MakeUnique<AtomicWaitExpr>(
+  return AppendExpr(std::make_unique<AtomicWaitExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
 Result BinaryReaderIR::OnAtomicFenceExpr(uint32_t consistency_model) {
-  return AppendExpr(MakeUnique<AtomicFenceExpr>(consistency_model));
+  return AppendExpr(std::make_unique<AtomicFenceExpr>(consistency_model));
 }
 
 Result BinaryReaderIR::OnAtomicNotifyExpr(Opcode opcode,
                                           Index memidx,
                                           Address alignment_log2,
                                           Address offset) {
-  return AppendExpr(MakeUnique<AtomicNotifyExpr>(
+  return AppendExpr(std::make_unique<AtomicNotifyExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
 Result BinaryReaderIR::OnBinaryExpr(Opcode opcode) {
-  return AppendExpr(MakeUnique<BinaryExpr>(opcode));
+  return AppendExpr(std::make_unique<BinaryExpr>(opcode));
 }
 
 Result BinaryReaderIR::OnBlockExpr(Type sig_type) {
-  auto expr = MakeUnique<BlockExpr>();
+  auto expr = std::make_unique<BlockExpr>();
   SetBlockDeclaration(&expr->block.decl, sig_type);
   ExprList* expr_list = &expr->block.exprs;
   CHECK_RESULT(AppendExpr(std::move(expr)));
-  PushLabel(LabelType::Block, expr_list);
-  return Result::Ok;
+  return PushLabel(LabelType::Block, expr_list);
 }
 
 Result BinaryReaderIR::OnBrExpr(Index depth) {
-  return AppendExpr(MakeUnique<BrExpr>(Var(depth, GetLocation())));
+  return AppendExpr(std::make_unique<BrExpr>(Var(depth, GetLocation())));
 }
 
 Result BinaryReaderIR::OnBrIfExpr(Index depth) {
-  return AppendExpr(MakeUnique<BrIfExpr>(Var(depth, GetLocation())));
+  return AppendExpr(std::make_unique<BrIfExpr>(Var(depth, GetLocation())));
 }
 
 Result BinaryReaderIR::OnBrTableExpr(Index num_targets,
                                      Index* target_depths,
                                      Index default_target_depth) {
-  auto expr = MakeUnique<BrTableExpr>();
+  auto expr = std::make_unique<BrTableExpr>();
   expr->default_target = Var(default_target_depth, GetLocation());
   expr->targets.resize(num_targets);
   for (Index i = 0; i < num_targets; ++i) {
@@ -831,42 +876,43 @@ Result BinaryReaderIR::OnBrTableExpr(Index num_targets,
 }
 
 Result BinaryReaderIR::OnCallExpr(Index func_index) {
-  return AppendExpr(MakeUnique<CallExpr>(Var(func_index, GetLocation())));
+  return AppendExpr(std::make_unique<CallExpr>(Var(func_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnCallIndirectExpr(Index sig_index, Index table_index) {
-  auto expr = MakeUnique<CallIndirectExpr>();
+  auto expr = std::make_unique<CallIndirectExpr>();
   SetFuncDeclaration(&expr->decl, Var(sig_index, GetLocation()));
   expr->table = Var(table_index, GetLocation());
   return AppendExpr(std::move(expr));
 }
 
 Result BinaryReaderIR::OnCallRefExpr() {
-  return AppendExpr(MakeUnique<CallRefExpr>());
+  return AppendExpr(std::make_unique<CallRefExpr>());
 }
 
 Result BinaryReaderIR::OnReturnCallExpr(Index func_index) {
-  return AppendExpr(MakeUnique<ReturnCallExpr>(Var(func_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<ReturnCallExpr>(Var(func_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnReturnCallIndirectExpr(Index sig_index,
                                                 Index table_index) {
-  auto expr = MakeUnique<ReturnCallIndirectExpr>();
+  auto expr = std::make_unique<ReturnCallIndirectExpr>();
   SetFuncDeclaration(&expr->decl, Var(sig_index, GetLocation()));
   expr->table = Var(table_index, GetLocation());
   return AppendExpr(std::move(expr));
 }
 
 Result BinaryReaderIR::OnCompareExpr(Opcode opcode) {
-  return AppendExpr(MakeUnique<CompareExpr>(opcode));
+  return AppendExpr(std::make_unique<CompareExpr>(opcode));
 }
 
 Result BinaryReaderIR::OnConvertExpr(Opcode opcode) {
-  return AppendExpr(MakeUnique<ConvertExpr>(opcode));
+  return AppendExpr(std::make_unique<ConvertExpr>(opcode));
 }
 
 Result BinaryReaderIR::OnDropExpr() {
-  return AppendExpr(MakeUnique<DropExpr>());
+  return AppendExpr(std::make_unique<DropExpr>());
 }
 
 Result BinaryReaderIR::OnElseExpr() {
@@ -921,186 +967,200 @@ Result BinaryReaderIR::OnEndExpr() {
 
 Result BinaryReaderIR::OnF32ConstExpr(uint32_t value_bits) {
   return AppendExpr(
-      MakeUnique<ConstExpr>(Const::F32(value_bits, GetLocation())));
+      std::make_unique<ConstExpr>(Const::F32(value_bits, GetLocation())));
 }
 
 Result BinaryReaderIR::OnF64ConstExpr(uint64_t value_bits) {
   return AppendExpr(
-      MakeUnique<ConstExpr>(Const::F64(value_bits, GetLocation())));
+      std::make_unique<ConstExpr>(Const::F64(value_bits, GetLocation())));
 }
 
 Result BinaryReaderIR::OnV128ConstExpr(v128 value_bits) {
   return AppendExpr(
-      MakeUnique<ConstExpr>(Const::V128(value_bits, GetLocation())));
+      std::make_unique<ConstExpr>(Const::V128(value_bits, GetLocation())));
 }
 
 Result BinaryReaderIR::OnGlobalGetExpr(Index global_index) {
   return AppendExpr(
-      MakeUnique<GlobalGetExpr>(Var(global_index, GetLocation())));
+      std::make_unique<GlobalGetExpr>(Var(global_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnLocalGetExpr(Index local_index) {
-  return AppendExpr(MakeUnique<LocalGetExpr>(Var(local_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<LocalGetExpr>(Var(local_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnI32ConstExpr(uint32_t value) {
-  return AppendExpr(MakeUnique<ConstExpr>(Const::I32(value, GetLocation())));
+  return AppendExpr(
+      std::make_unique<ConstExpr>(Const::I32(value, GetLocation())));
 }
 
 Result BinaryReaderIR::OnI64ConstExpr(uint64_t value) {
-  return AppendExpr(MakeUnique<ConstExpr>(Const::I64(value, GetLocation())));
+  return AppendExpr(
+      std::make_unique<ConstExpr>(Const::I64(value, GetLocation())));
 }
 
 Result BinaryReaderIR::OnIfExpr(Type sig_type) {
-  auto expr = MakeUnique<IfExpr>();
+  auto expr = std::make_unique<IfExpr>();
   SetBlockDeclaration(&expr->true_.decl, sig_type);
   ExprList* expr_list = &expr->true_.exprs;
   CHECK_RESULT(AppendExpr(std::move(expr)));
-  PushLabel(LabelType::If, expr_list);
-  return Result::Ok;
+  return PushLabel(LabelType::If, expr_list);
 }
 
 Result BinaryReaderIR::OnLoadExpr(Opcode opcode,
                                   Index memidx,
                                   Address alignment_log2,
                                   Address offset) {
-  return AppendExpr(MakeUnique<LoadExpr>(opcode, Var(memidx, GetLocation()),
-                                         1 << alignment_log2, offset));
+  return AppendExpr(std::make_unique<LoadExpr>(
+      opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
 Result BinaryReaderIR::OnLoopExpr(Type sig_type) {
-  auto expr = MakeUnique<LoopExpr>();
+  auto expr = std::make_unique<LoopExpr>();
   SetBlockDeclaration(&expr->block.decl, sig_type);
   ExprList* expr_list = &expr->block.exprs;
   CHECK_RESULT(AppendExpr(std::move(expr)));
-  PushLabel(LabelType::Loop, expr_list);
-  return Result::Ok;
+  return PushLabel(LabelType::Loop, expr_list);
 }
 
 Result BinaryReaderIR::OnMemoryCopyExpr(Index srcmemidx, Index destmemidx) {
-  return AppendExpr(MakeUnique<MemoryCopyExpr>(Var(srcmemidx, GetLocation()),
-                                               Var(destmemidx, GetLocation())));
+  return AppendExpr(std::make_unique<MemoryCopyExpr>(
+      Var(srcmemidx, GetLocation()), Var(destmemidx, GetLocation())));
 }
 
 Result BinaryReaderIR::OnDataDropExpr(Index segment) {
-  return AppendExpr(MakeUnique<DataDropExpr>(Var(segment, GetLocation())));
+  return AppendExpr(
+      std::make_unique<DataDropExpr>(Var(segment, GetLocation())));
 }
 
 Result BinaryReaderIR::OnMemoryFillExpr(Index memidx) {
-  return AppendExpr(MakeUnique<MemoryFillExpr>(Var(memidx, GetLocation())));
+  return AppendExpr(
+      std::make_unique<MemoryFillExpr>(Var(memidx, GetLocation())));
 }
 
 Result BinaryReaderIR::OnMemoryGrowExpr(Index memidx) {
-  return AppendExpr(MakeUnique<MemoryGrowExpr>(Var(memidx, GetLocation())));
+  return AppendExpr(
+      std::make_unique<MemoryGrowExpr>(Var(memidx, GetLocation())));
 }
 
 Result BinaryReaderIR::OnMemoryInitExpr(Index segment, Index memidx) {
-  return AppendExpr(MakeUnique<MemoryInitExpr>(Var(segment, GetLocation()),
-                                               Var(memidx, GetLocation())));
+  return AppendExpr(std::make_unique<MemoryInitExpr>(
+      Var(segment, GetLocation()), Var(memidx, GetLocation())));
 }
 
 Result BinaryReaderIR::OnMemorySizeExpr(Index memidx) {
-  return AppendExpr(MakeUnique<MemorySizeExpr>(Var(memidx, GetLocation())));
+  return AppendExpr(
+      std::make_unique<MemorySizeExpr>(Var(memidx, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableCopyExpr(Index dst_index, Index src_index) {
-  return AppendExpr(MakeUnique<TableCopyExpr>(Var(dst_index, GetLocation()),
-                                              Var(src_index, GetLocation())));
+  return AppendExpr(std::make_unique<TableCopyExpr>(
+      Var(dst_index, GetLocation()), Var(src_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnElemDropExpr(Index segment) {
-  return AppendExpr(MakeUnique<ElemDropExpr>(Var(segment, GetLocation())));
+  return AppendExpr(
+      std::make_unique<ElemDropExpr>(Var(segment, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableInitExpr(Index segment, Index table_index) {
-  return AppendExpr(MakeUnique<TableInitExpr>(Var(segment, GetLocation()),
-                                              Var(table_index, GetLocation())));
+  return AppendExpr(std::make_unique<TableInitExpr>(
+      Var(segment, GetLocation()), Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableGetExpr(Index table_index) {
-  return AppendExpr(MakeUnique<TableGetExpr>(Var(table_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<TableGetExpr>(Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableSetExpr(Index table_index) {
-  return AppendExpr(MakeUnique<TableSetExpr>(Var(table_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<TableSetExpr>(Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableGrowExpr(Index table_index) {
-  return AppendExpr(MakeUnique<TableGrowExpr>(Var(table_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<TableGrowExpr>(Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableSizeExpr(Index table_index) {
-  return AppendExpr(MakeUnique<TableSizeExpr>(Var(table_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<TableSizeExpr>(Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTableFillExpr(Index table_index) {
-  return AppendExpr(MakeUnique<TableFillExpr>(Var(table_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<TableFillExpr>(Var(table_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnRefFuncExpr(Index func_index) {
-  return AppendExpr(MakeUnique<RefFuncExpr>(Var(func_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<RefFuncExpr>(Var(func_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnRefNullExpr(Type type) {
-  return AppendExpr(MakeUnique<RefNullExpr>(type));
+  return AppendExpr(std::make_unique<RefNullExpr>(type));
 }
 
 Result BinaryReaderIR::OnRefIsNullExpr() {
-  return AppendExpr(MakeUnique<RefIsNullExpr>());
+  return AppendExpr(std::make_unique<RefIsNullExpr>());
 }
 
 Result BinaryReaderIR::OnNopExpr() {
-  return AppendExpr(MakeUnique<NopExpr>());
+  return AppendExpr(std::make_unique<NopExpr>());
 }
 
 Result BinaryReaderIR::OnRethrowExpr(Index depth) {
-  return AppendExpr(MakeUnique<RethrowExpr>(Var(depth, GetLocation())));
+  return AppendExpr(std::make_unique<RethrowExpr>(Var(depth, GetLocation())));
 }
 
 Result BinaryReaderIR::OnReturnExpr() {
-  return AppendExpr(MakeUnique<ReturnExpr>());
+  return AppendExpr(std::make_unique<ReturnExpr>());
 }
 
 Result BinaryReaderIR::OnSelectExpr(Index result_count, Type* result_types) {
   TypeVector results;
   results.assign(result_types, result_types + result_count);
-  return AppendExpr(MakeUnique<SelectExpr>(results));
+  return AppendExpr(std::make_unique<SelectExpr>(results));
 }
 
 Result BinaryReaderIR::OnGlobalSetExpr(Index global_index) {
   return AppendExpr(
-      MakeUnique<GlobalSetExpr>(Var(global_index, GetLocation())));
+      std::make_unique<GlobalSetExpr>(Var(global_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnLocalSetExpr(Index local_index) {
-  return AppendExpr(MakeUnique<LocalSetExpr>(Var(local_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<LocalSetExpr>(Var(local_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnStoreExpr(Opcode opcode,
                                    Index memidx,
                                    Address alignment_log2,
                                    Address offset) {
-  return AppendExpr(MakeUnique<StoreExpr>(opcode, Var(memidx, GetLocation()),
-                                          1 << alignment_log2, offset));
+  return AppendExpr(std::make_unique<StoreExpr>(
+      opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
 Result BinaryReaderIR::OnThrowExpr(Index tag_index) {
-  return AppendExpr(MakeUnique<ThrowExpr>(Var(tag_index, GetLocation())));
+  return AppendExpr(std::make_unique<ThrowExpr>(Var(tag_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnLocalTeeExpr(Index local_index) {
-  return AppendExpr(MakeUnique<LocalTeeExpr>(Var(local_index, GetLocation())));
+  return AppendExpr(
+      std::make_unique<LocalTeeExpr>(Var(local_index, GetLocation())));
 }
 
 Result BinaryReaderIR::OnTryExpr(Type sig_type) {
-  auto expr_ptr = MakeUnique<TryExpr>();
+  auto expr_ptr = std::make_unique<TryExpr>();
   // Save expr so it can be used below, after expr_ptr has been moved.
   TryExpr* expr = expr_ptr.get();
   ExprList* expr_list = &expr->block.exprs;
   SetBlockDeclaration(&expr->block.decl, sig_type);
   CHECK_RESULT(AppendExpr(std::move(expr_ptr)));
-  PushLabel(LabelType::Try, expr_list, expr);
-  return Result::Ok;
+  module_->features_used.exceptions = true;
+  return PushLabel(LabelType::Try, expr_list, expr);
 }
 
 Result BinaryReaderIR::AppendCatch(Catch&& catch_) {
@@ -1165,24 +1225,28 @@ Result BinaryReaderIR::OnDelegateExpr(Index depth) {
 }
 
 Result BinaryReaderIR::OnUnaryExpr(Opcode opcode) {
-  return AppendExpr(MakeUnique<UnaryExpr>(opcode));
+  return AppendExpr(std::make_unique<UnaryExpr>(opcode));
 }
 
 Result BinaryReaderIR::OnTernaryExpr(Opcode opcode) {
-  return AppendExpr(MakeUnique<TernaryExpr>(opcode));
+  return AppendExpr(std::make_unique<TernaryExpr>(opcode));
 }
 
 Result BinaryReaderIR::OnUnreachableExpr() {
-  return AppendExpr(MakeUnique<UnreachableExpr>());
+  return AppendExpr(std::make_unique<UnreachableExpr>());
 }
 
 Result BinaryReaderIR::EndFunctionBody(Index index) {
   current_func_ = nullptr;
+  if (!label_stack_.empty()) {
+    PrintError("function %" PRIindex " missing end marker", index);
+    return Result::Error;
+  }
   return Result::Ok;
 }
 
 Result BinaryReaderIR::OnSimdLaneOpExpr(Opcode opcode, uint64_t value) {
-  return AppendExpr(MakeUnique<SimdLaneOpExpr>(opcode, value));
+  return AppendExpr(std::make_unique<SimdLaneOpExpr>(opcode, value));
 }
 
 Result BinaryReaderIR::OnSimdLoadLaneExpr(Opcode opcode,
@@ -1190,7 +1254,7 @@ Result BinaryReaderIR::OnSimdLoadLaneExpr(Opcode opcode,
                                           Address alignment_log2,
                                           Address offset,
                                           uint64_t value) {
-  return AppendExpr(MakeUnique<SimdLoadLaneExpr>(
+  return AppendExpr(std::make_unique<SimdLoadLaneExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset, value));
 }
 
@@ -1199,19 +1263,19 @@ Result BinaryReaderIR::OnSimdStoreLaneExpr(Opcode opcode,
                                            Address alignment_log2,
                                            Address offset,
                                            uint64_t value) {
-  return AppendExpr(MakeUnique<SimdStoreLaneExpr>(
+  return AppendExpr(std::make_unique<SimdStoreLaneExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset, value));
 }
 
 Result BinaryReaderIR::OnSimdShuffleOpExpr(Opcode opcode, v128 value) {
-  return AppendExpr(MakeUnique<SimdShuffleOpExpr>(opcode, value));
+  return AppendExpr(std::make_unique<SimdShuffleOpExpr>(opcode, value));
 }
 
 Result BinaryReaderIR::OnLoadSplatExpr(Opcode opcode,
                                        Index memidx,
                                        Address alignment_log2,
                                        Address offset) {
-  return AppendExpr(MakeUnique<LoadSplatExpr>(
+  return AppendExpr(std::make_unique<LoadSplatExpr>(
       opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
@@ -1219,8 +1283,8 @@ Result BinaryReaderIR::OnLoadZeroExpr(Opcode opcode,
                                       Index memidx,
                                       Address alignment_log2,
                                       Address offset) {
-  return AppendExpr(MakeUnique<LoadZeroExpr>(opcode, Var(memidx, GetLocation()),
-                                             1 << alignment_log2, offset));
+  return AppendExpr(std::make_unique<LoadZeroExpr>(
+      opcode, Var(memidx, GetLocation()), 1 << alignment_log2, offset));
 }
 
 Result BinaryReaderIR::OnElemSegmentCount(Index count) {
@@ -1233,7 +1297,7 @@ Result BinaryReaderIR::OnElemSegmentCount(Index count) {
 Result BinaryReaderIR::BeginElemSegment(Index index,
                                         Index table_index,
                                         uint8_t flags) {
-  auto field = MakeUnique<ElemSegmentModuleField>(GetLocation());
+  auto field = std::make_unique<ElemSegmentModuleField>(GetLocation());
   ElemSegment& elem_segment = field->elem_segment;
   elem_segment.table_var = Var(table_index, GetLocation());
   if ((flags & SegDeclared) == SegDeclared) {
@@ -1248,8 +1312,7 @@ Result BinaryReaderIR::BeginElemSegment(Index index,
 }
 
 Result BinaryReaderIR::BeginInitExpr(ExprList* expr) {
-  PushLabel(LabelType::InitExpr, expr);
-  return Result::Ok;
+  return PushLabel(LabelType::InitExpr, expr);
 }
 
 Result BinaryReaderIR::BeginElemSegmentInitExpr(Index index) {
@@ -1259,6 +1322,10 @@ Result BinaryReaderIR::BeginElemSegmentInitExpr(Index index) {
 }
 
 Result BinaryReaderIR::EndInitExpr() {
+  if (!label_stack_.empty()) {
+    PrintError("init expression missing end marker");
+    return Result::Error;
+  }
   return Result::Ok;
 }
 
@@ -1288,7 +1355,7 @@ Result BinaryReaderIR::OnElemSegmentElemExpr_RefNull(Index segment_index,
   ElemSegment* segment = module_->elem_segments[segment_index];
   Location loc = GetLocation();
   ExprList init_expr;
-  init_expr.push_back(MakeUnique<RefNullExpr>(type, loc));
+  init_expr.push_back(std::make_unique<RefNullExpr>(type, loc));
   segment->elem_exprs.push_back(std::move(init_expr));
   return Result::Ok;
 }
@@ -1299,7 +1366,7 @@ Result BinaryReaderIR::OnElemSegmentElemExpr_RefFunc(Index segment_index,
   ElemSegment* segment = module_->elem_segments[segment_index];
   Location loc = GetLocation();
   ExprList init_expr;
-  init_expr.push_back(MakeUnique<RefFuncExpr>(Var(func_index, loc), loc));
+  init_expr.push_back(std::make_unique<RefFuncExpr>(Var(func_index, loc), loc));
   segment->elem_exprs.push_back(std::move(init_expr));
   return Result::Ok;
 }
@@ -1314,7 +1381,7 @@ Result BinaryReaderIR::OnDataSegmentCount(Index count) {
 Result BinaryReaderIR::BeginDataSegment(Index index,
                                         Index memory_index,
                                         uint8_t flags) {
-  auto field = MakeUnique<DataSegmentModuleField>(GetLocation());
+  auto field = std::make_unique<DataSegmentModuleField>(GetLocation());
   DataSegment& data_segment = field->data_segment;
   data_segment.memory_var = Var(memory_index, GetLocation());
   if ((flags & SegPassive) == SegPassive) {
@@ -1572,8 +1639,8 @@ Result BinaryReaderIR::OnCodeMetadata(Offset offset,
                                       Address size) {
   std::vector<uint8_t> data_(static_cast<const uint8_t*>(data),
                              static_cast<const uint8_t*>(data) + size);
-  auto meta =
-      MakeUnique<CodeMetadataExpr>(current_metadata_name_, std::move(data_));
+  auto meta = std::make_unique<CodeMetadataExpr>(current_metadata_name_,
+                                                 std::move(data_));
   meta->loc.offset = offset;
   code_metadata_queue_.push_metadata(std::move(meta));
   return Result::Ok;
@@ -1593,10 +1660,11 @@ Result BinaryReaderIR::OnLocalName(Index func_index,
 }
 
 Result BinaryReaderIR::OnTagType(Index index, Index sig_index) {
-  auto field = MakeUnique<TagModuleField>(GetLocation());
+  auto field = std::make_unique<TagModuleField>(GetLocation());
   Tag& tag = field->tag;
   SetFuncDeclaration(&tag.decl, Var(sig_index, GetLocation()));
   module_->AppendField(std::move(field));
+  module_->features_used.exceptions = true;
   return Result::Ok;
 }
 
