@@ -15,17 +15,33 @@
 
 #define UNREACHABLE TRAP(UNREACHABLE)
 
-#define CALL_INDIRECT(table, t, ft, x, ...)             \
-  (LIKELY((x) < table.size && table.data[x].func &&     \
-          table.data[x].func_type == func_types[ft]) || \
-       TRAP(CALL_INDIRECT),                             \
+static inline bool func_types_eq(const wasm_rt_func_type_t a,
+                                 const wasm_rt_func_type_t b) {
+  return (a == b) || LIKELY(a && b && !memcmp(a, b, 32));
+}
+
+#define CALL_INDIRECT(table, t, ft, x, ...)              \
+  (LIKELY((x) < table.size && table.data[x].func &&      \
+          func_types_eq(ft, table.data[x].func_type)) || \
+       TRAP(CALL_INDIRECT),                              \
    ((t)table.data[x].func)(__VA_ARGS__))
 
+#ifdef SUPPORT_MEMORY64
+#define RANGE_CHECK(mem, offset, len)              \
+  do {                                             \
+    uint64_t res;                                  \
+    if (__builtin_add_overflow(offset, len, &res)) \
+      TRAP(OOB);                                   \
+    if (UNLIKELY(res > mem->size))                 \
+      TRAP(OOB);                                   \
+  } while (0);
+#else
 #define RANGE_CHECK(mem, offset, len)               \
   if (UNLIKELY(offset + (uint64_t)len > mem->size)) \
     TRAP(OOB);
+#endif
 
-#if WASM_RT_MEMCHECK_SIGNAL_HANDLER
+#if WASM_RT_MEMCHECK_GUARD_PAGES
 #define MEMCHECK(mem, a, t)
 #else
 #define MEMCHECK(mem, a, t) RANGE_CHECK(mem, a, sizeof(t))
@@ -39,6 +55,9 @@
 
 #if WABT_BIG_ENDIAN
 static inline void load_data(void* dest, const void* src, size_t n) {
+  if (!n) {
+    return;
+  }
   size_t i = 0;
   u8* dest_chars = dest;
   memcpy(dest, src, n);
@@ -72,6 +91,9 @@ static inline void load_data(void* dest, const void* src, size_t n) {
   }
 #else
 static inline void load_data(void* dest, const void* src, size_t n) {
+  if (!n) {
+    return;
+  }
   memcpy(dest, src, n);
 }
 #define LOAD_DATA(m, o, i, s)      \
@@ -410,7 +432,7 @@ static float wasm_fabsf(float x) {
   if (UNLIKELY(isnan(x))) {
     uint32_t tmp;
     memcpy(&tmp, &x, 4);
-    tmp = tmp & ~(1 << 31);
+    tmp = tmp & ~(1UL << 31);
     memcpy(&x, &tmp, 4);
     return x;
   }
@@ -421,7 +443,7 @@ static double wasm_fabs(double x) {
   if (UNLIKELY(isnan(x))) {
     uint64_t tmp;
     memcpy(&tmp, &x, 8);
-    tmp = tmp & ~(1ll << 63);
+    tmp = tmp & ~(1ULL << 63);
     memcpy(&x, &tmp, 8);
     return x;
   }
@@ -469,7 +491,7 @@ static inline void memory_init(wasm_rt_memory_t* dest,
 }
 
 typedef struct {
-  uint32_t func_type_index;
+  wasm_rt_func_type_t type;
   wasm_rt_function_ptr_t func;
   size_t module_offset;
 } wasm_elem_segment_expr_t;
@@ -480,17 +502,16 @@ static inline void funcref_table_init(wasm_rt_funcref_table_t* dest,
                                       u32 dest_addr,
                                       u32 src_addr,
                                       u32 n,
-                                      void* module_instance,
-                                      const u32* func_types) {
+                                      void* module_instance) {
   if (UNLIKELY(src_addr + (uint64_t)n > src_size))
     TRAP(OOB);
   if (UNLIKELY(dest_addr + (uint64_t)n > dest->size))
     TRAP(OOB);
   for (u32 i = 0; i < n; i++) {
     const wasm_elem_segment_expr_t* src_expr = &src[src_addr + i];
-    dest->data[dest_addr + i] = (wasm_rt_funcref_t){
-        func_types[src_expr->func_type_index], src_expr->func,
-        (char*)module_instance + src_expr->module_offset};
+    dest->data[dest_addr + i] =
+        (wasm_rt_funcref_t){src_expr->type, src_expr->func,
+                            (char*)module_instance + src_expr->module_offset};
   }
 }
 
@@ -561,4 +582,12 @@ DEFINE_TABLE_SET(externref)
 DEFINE_TABLE_FILL(funcref)
 DEFINE_TABLE_FILL(externref)
 
-static bool s_module_initialized = false;
+#if defined(__GNUC__) || defined(__clang__)
+#define FUNC_TYPE_DECL_EXTERN_T(x) extern const char* const x
+#define FUNC_TYPE_EXTERN_T(x) const char* const x
+#define FUNC_TYPE_T(x) static const char* const x
+#else
+#define FUNC_TYPE_DECL_EXTERN_T(x) extern const char x[]
+#define FUNC_TYPE_EXTERN_T(x) const char x[]
+#define FUNC_TYPE_T(x) static const char x[]
+#endif
