@@ -24,6 +24,7 @@
 #include <string>
 #include <vector>
 
+#include "wabt/binary-reader-ir.h"
 #include "wabt/binary-reader.h"
 #include "wabt/cast.h"
 #include "wabt/common.h"
@@ -189,9 +190,8 @@ int LaneCountFromType(Type type) {
 }
 
 ExpectedValue GetLane(const ExpectedValue& ev, int lane) {
-  int lane_count = LaneCountFromType(ev.lane_type);
   assert(ev.value.type == Type::V128);
-  assert(lane < lane_count);
+  assert(lane < LaneCountFromType(ev.lane_type));
 
   ExpectedValue result;
   result.value.type = ev.lane_type;
@@ -236,9 +236,8 @@ ExpectedValue GetLane(const ExpectedValue& ev, int lane) {
 }
 
 TypedValue GetLane(const TypedValue& tv, Type lane_type, int lane) {
-  int lane_count = LaneCountFromType(lane_type);
   assert(tv.type == Type::V128);
-  assert(lane < lane_count);
+  assert(lane < LaneCountFromType(lane_type));
 
   TypedValue result;
   result.type = lane_type;
@@ -276,6 +275,30 @@ TypedValue GetLane(const TypedValue& tv, Type lane_type, int lane) {
   return result;
 }
 
+bool ValidIR(const std::string& filename) {
+  std::vector<uint8_t> file_data;
+
+  if (Failed(ReadFile(filename, &file_data))) {
+    return false;
+  }
+
+  const bool kReadDebugNames = true;
+  const bool kStopOnFirstError = true;
+  const bool kFailOnCustomSectionError = true;
+  ReadBinaryOptions options(s_features, s_log_stream.get(), kReadDebugNames,
+                            kStopOnFirstError, kFailOnCustomSectionError);
+
+  Errors errors;
+  wabt::Module module;
+  if (Failed(ReadBinaryIr(filename.c_str(), file_data.data(), file_data.size(),
+                          options, &errors, &module))) {
+    return false;
+  }
+
+  return Succeeded(
+      ValidateModule(&module, &errors, ValidateOptions{s_features}));
+}
+
 class AssertReturnCommand : public CommandMixin<CommandType::AssertReturn> {
  public:
   Action action;
@@ -1004,7 +1027,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
   EXPECT("{");
   EXPECT_KEY("type");
   if (Match("\"module\"")) {
-    auto command = MakeUnique<ModuleCommand>();
+    auto command = std::make_unique<ModuleCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1012,7 +1035,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseFilename(&command->filename));
     *out_command = std::move(command);
   } else if (Match("\"action\"")) {
-    auto command = MakeUnique<ActionCommand>();
+    auto command = std::make_unique<ActionCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1021,7 +1044,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseActionResult());
     *out_command = std::move(command);
   } else if (Match("\"register\"")) {
-    auto command = MakeUnique<RegisterCommand>();
+    auto command = std::make_unique<RegisterCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1029,7 +1052,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     PARSE_KEY_STRING_VALUE("as", &command->as);
     *out_command = std::move(command);
   } else if (Match("\"assert_malformed\"")) {
-    auto command = MakeUnique<AssertMalformedCommand>();
+    auto command = std::make_unique<AssertMalformedCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1040,7 +1063,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseModuleType(&command->type));
     *out_command = std::move(command);
   } else if (Match("\"assert_invalid\"")) {
-    auto command = MakeUnique<AssertInvalidCommand>();
+    auto command = std::make_unique<AssertInvalidCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1051,7 +1074,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseModuleType(&command->type));
     *out_command = std::move(command);
   } else if (Match("\"assert_unlinkable\"")) {
-    auto command = MakeUnique<AssertUnlinkableCommand>();
+    auto command = std::make_unique<AssertUnlinkableCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1062,7 +1085,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseModuleType(&command->type));
     *out_command = std::move(command);
   } else if (Match("\"assert_uninstantiable\"")) {
-    auto command = MakeUnique<AssertUninstantiableCommand>();
+    auto command = std::make_unique<AssertUninstantiableCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1073,7 +1096,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseModuleType(&command->type));
     *out_command = std::move(command);
   } else if (Match("\"assert_return\"")) {
-    auto command = MakeUnique<AssertReturnCommand>();
+    auto command = std::make_unique<AssertReturnCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1088,7 +1111,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseExpectedValues(&command->expected));
     *out_command = std::move(command);
   } else if (Match("\"assert_trap\"")) {
-    auto command = MakeUnique<AssertTrapCommand>();
+    auto command = std::make_unique<AssertTrapCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1099,7 +1122,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
     CHECK_RESULT(ParseActionResult());
     *out_command = std::move(command);
   } else if (Match("\"assert_exhaustion\"")) {
-    auto command = MakeUnique<AssertExhaustionCommand>();
+    auto command = std::make_unique<AssertExhaustionCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1114,7 +1137,7 @@ wabt::Result JSONParser::ParseCommand(CommandPtr* out_command) {
       PrintError("invalid command: exceptions not allowed");
       return wabt::Result::Error;
     }
-    auto command = MakeUnique<AssertExceptionCommand>();
+    auto command = std::make_unique<AssertExceptionCommand>();
     EXPECT(",");
     CHECK_RESULT(ParseLine(&command->line));
     EXPECT(",");
@@ -1500,6 +1523,12 @@ wabt::Result CommandRunner::OnModuleCommand(const ModuleCommand* command) {
     return wabt::Result::Error;
   }
 
+  if (!ValidIR(command->filename)) {
+    PrintError(command->line, "IR Validator thinks module is invalid: \"%s\"",
+               command->filename.c_str());
+    return wabt::Result::Error;
+  }
+
   RefVec imports;
   PopulateImports(module, &imports);
 
@@ -1543,6 +1572,12 @@ wabt::Result CommandRunner::OnAssertMalformedCommand(
     return wabt::Result::Error;
   }
 
+  if (ValidIR(command->filename)) {
+    PrintError(command->line, "IR Validator thinks module is valid: \"%s\"",
+               command->filename.c_str());
+    return wabt::Result::Error;
+  }
+
   return wabt::Result::Ok;
 }
 
@@ -1572,6 +1607,12 @@ wabt::Result CommandRunner::OnAssertUnlinkableCommand(
     return wabt::Result::Error;
   }
 
+  if (!ValidIR(command->filename)) {
+    PrintError(command->line, "IR Validator thinks module is invalid: \"%s\"",
+               command->filename.c_str());
+    return wabt::Result::Error;
+  }
+
   RefVec imports;
   PopulateImports(module, &imports);
 
@@ -1599,6 +1640,12 @@ wabt::Result CommandRunner::OnAssertInvalidCommand(
     return wabt::Result::Error;
   }
 
+  if (ValidIR(command->filename)) {
+    PrintError(command->line, "IR Validator thinks module is valid: \"%s\"",
+               command->filename.c_str());
+    return wabt::Result::Error;
+  }
+
   return wabt::Result::Ok;
 }
 
@@ -1613,6 +1660,12 @@ wabt::Result CommandRunner::OnAssertUninstantiableCommand(
     return wabt::Result::Error;
   }
 
+  if (!ValidIR(command->filename)) {
+    PrintError(command->line, "IR Validator thinks module is invalid: \"%s\"",
+               command->filename.c_str());
+    return wabt::Result::Error;
+  }
+
   RefVec imports;
   PopulateImports(module, &imports);
 
