@@ -28,6 +28,7 @@
 #include "wabt/generate-names.h"
 #include "wabt/ir.h"
 #include "wabt/option-parser.h"
+#include "wabt/result.h"
 #include "wabt/stream.h"
 #include "wabt/validator.h"
 #include "wabt/wast-lexer.h"
@@ -39,6 +40,7 @@ using namespace wabt;
 static int s_verbose;
 static std::string s_infile;
 static std::string s_outfile;
+static unsigned int s_num_outputs = 1;
 static Features s_features;
 static WriteCOptions s_write_c_options;
 static bool s_read_debug_names = true;
@@ -57,8 +59,8 @@ examples:
 )";
 
 static const std::string supported_features[] = {
-    "multi-memory", "multi-value", "sign-extend", "saturating-float-to-int",
-    "exceptions"};
+    "multi-memory", "multi-value", "sign-extension", "saturating-float-to-int",
+    "exceptions",   "memory64",    "extended-const", "simd"};
 
 static bool IsFeatureSupported(const std::string& feature) {
   return std::find(std::begin(supported_features), std::end(supported_features),
@@ -79,6 +81,9 @@ static void ParseOptions(int argc, char** argv) {
         s_outfile = argument;
         ConvertBackslashToSlash(&s_outfile);
       });
+  parser.AddOption(
+      '\0', "num-outputs", "NUM", "Number of output files to write",
+      [](const char* argument) { s_num_outputs = atoi(argument); });
   parser.AddOption(
       'n', "module-name", "MODNAME",
       "Unique name for the module being generated. This name is prefixed to\n"
@@ -100,7 +105,7 @@ static void ParseOptions(int argc, char** argv) {
 #define WABT_FEATURE(variable, flag, default_, help)   \
   any_non_supported_feature |=                         \
       (s_features.variable##_enabled() != default_) && \
-      !IsFeatureSupported(flag);
+      s_features.variable##_enabled() && !IsFeatureSupported(flag);
 #include "wabt/feature.def"
 #undef WABT_FEATURE
 
@@ -121,66 +126,86 @@ static std::string_view strip_extension(std::string_view s) {
   return result;
 }
 
+Result Wasm2cMain(Errors& errors) {
+  if (s_num_outputs < 1) {
+    fprintf(stderr, "Number of output files must be positive.\n");
+    return Result::Error;
+  }
+
+  std::vector<uint8_t> file_data;
+  CHECK_RESULT(ReadFile(s_infile.c_str(), &file_data));
+
+  Module module;
+  const bool kStopOnFirstError = true;
+  const bool kFailOnCustomSectionError = true;
+  ReadBinaryOptions options(s_features, s_log_stream.get(), s_read_debug_names,
+                            kStopOnFirstError, kFailOnCustomSectionError);
+  CHECK_RESULT(ReadBinaryIr(s_infile.c_str(), file_data.data(),
+                            file_data.size(), options, &errors, &module));
+  CHECK_RESULT(ValidateModule(&module, &errors, s_features));
+  CHECK_RESULT(GenerateNames(&module));
+  /* TODO(binji): This shouldn't fail; if a name can't be applied
+   * (because the index is invalid, say) it should just be skipped. */
+  ApplyNames(&module);
+
+  if (!s_outfile.empty()) {
+    std::string header_name_full =
+        std::string(strip_extension(s_outfile)) + ".h";
+    std::vector<FileStream> c_streams;
+    if (s_num_outputs == 1) {
+      c_streams.emplace_back(s_outfile.c_str());
+    } else {
+      std::string output_prefix{strip_extension(s_outfile)};
+      for (unsigned int i = 0; i < s_num_outputs; i++) {
+        c_streams.emplace_back(output_prefix + "_" + std::to_string(i) + ".c");
+      }
+    }
+    std::vector<Stream*> c_stream_ptrs;
+    for (auto& s : c_streams) {
+      c_stream_ptrs.emplace_back(&s);
+    }
+    FileStream h_stream(header_name_full);
+    std::string_view header_name = GetBasename(header_name_full);
+    if (s_write_c_options.module_name.empty()) {
+      s_write_c_options.module_name = module.name;
+      if (s_write_c_options.module_name.empty()) {
+        // In the absence of module name in the names section use the filename.
+        s_write_c_options.module_name = StripExtension(GetBasename(s_infile));
+      }
+    }
+    if (s_num_outputs == 1) {
+      CHECK_RESULT(WriteC(std::move(c_stream_ptrs), &h_stream, c_stream_ptrs[0],
+                          std::string(header_name).c_str(), "", &module,
+                          s_write_c_options));
+    } else {
+      std::string header_impl_name_full =
+          std::string(strip_extension(s_outfile)) + "-impl.h";
+      FileStream h_impl_stream(header_impl_name_full);
+      std::string_view header_impl_name = GetBasename(header_impl_name_full);
+      CHECK_RESULT(WriteC(std::move(c_stream_ptrs), &h_stream, &h_impl_stream,
+                          std::string(header_name).c_str(),
+                          std::string(header_impl_name).c_str(), &module,
+                          s_write_c_options));
+    }
+  } else {
+    FileStream stream(stdout);
+    CHECK_RESULT(WriteC({&stream}, &stream, &stream, "wasm.h", "", &module,
+                        s_write_c_options));
+  }
+
+  return Result::Ok;
+}
+
 int ProgramMain(int argc, char** argv) {
   Result result;
 
   InitStdio();
   ParseOptions(argc, argv);
 
-  std::vector<uint8_t> file_data;
-  result = ReadFile(s_infile.c_str(), &file_data);
-  if (Succeeded(result)) {
-    Errors errors;
-    Module module;
-    const bool kStopOnFirstError = true;
-    const bool kFailOnCustomSectionError = true;
-    ReadBinaryOptions options(s_features, s_log_stream.get(),
-                              s_read_debug_names, kStopOnFirstError,
-                              kFailOnCustomSectionError);
-    result = ReadBinaryIr(s_infile.c_str(), file_data.data(), file_data.size(),
-                          options, &errors, &module);
-    if (Succeeded(result)) {
-      if (Succeeded(result)) {
-        ValidateOptions options(s_features);
-        result = ValidateModule(&module, &errors, options);
-        result |= GenerateNames(&module);
-      }
+  Errors errors;
+  result = Wasm2cMain(errors);
+  FormatErrorsToFile(errors, Location::Type::Binary);
 
-      if (Succeeded(result)) {
-        /* TODO(binji): This shouldn't fail; if a name can't be applied
-         * (because the index is invalid, say) it should just be skipped. */
-        Result dummy_result = ApplyNames(&module);
-        WABT_USE(dummy_result);
-      }
-
-      if (Succeeded(result)) {
-        if (!s_outfile.empty()) {
-          std::string header_name_full =
-              std::string(strip_extension(s_outfile)) + ".h";
-          FileStream c_stream(s_outfile.c_str());
-          FileStream h_stream(header_name_full);
-          std::string_view header_name = GetBasename(header_name_full);
-          if (s_write_c_options.module_name.empty()) {
-            s_write_c_options.module_name = module.name;
-            if (s_write_c_options.module_name.empty()) {
-              // In the absence of module name in the names section use the
-              // filename.
-              s_write_c_options.module_name =
-                  StripExtension(GetBasename(s_infile));
-            }
-          }
-          result =
-              WriteC(&c_stream, &h_stream, std::string(header_name).c_str(),
-                     &module, s_write_c_options);
-        } else {
-          FileStream stream(stdout);
-          result =
-              WriteC(&stream, &stream, "wasm.h", &module, s_write_c_options);
-        }
-      }
-    }
-    FormatErrorsToFile(errors, Location::Type::Binary);
-  }
   return result != Result::Ok;
 }
 
