@@ -31,9 +31,11 @@ from utils import Error
 
 SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
 WASM2C_DIR = os.path.join(find_exe.REPO_ROOT_DIR, 'wasm2c')
+SIMDE_DIR = os.path.join(find_exe.REPO_ROOT_DIR, 'third_party/simde')
 IS_WINDOWS = sys.platform == 'win32'
 IS_MACOS = platform.mac_ver()[0] != ''
 MAX_COMMANDS_PER_FUNCTION = 1024  # GCC has trouble with extremely long function bodies
+SKIPPED = 3
 
 
 def ReinterpretF32(f32_bits):
@@ -86,7 +88,7 @@ def F64ToC(f64_bits):
 
 
 def MangleType(t):
-    return {'i32': 'i', 'i64': 'j', 'f32': 'f', 'f64': 'd',
+    return {'i32': 'i', 'i64': 'j', 'f32': 'f', 'f64': 'd', 'v128': 'o',
             'externref': 'e', 'funcref': 'r'}[t]
 
 
@@ -99,11 +101,19 @@ def MangleTypes(types):
 def MangleName(s):
     def Mangle(match):
         s = match.group(0)
-        return b'Z%02X' % s[0]
+        return b'0x%02X' % s[0]
 
-    # NOTE(binji): Z is not allowed.
-    pattern = b'([^_a-zA-Y0-9])'
-    return 'Z_' + re.sub(pattern, Mangle, s.encode('utf-8')).decode('utf-8')
+    # escape underscores at beginning and end
+    s = re.sub(b'((^_)|(_$))', Mangle, s.encode('utf-8'))
+
+    # NOTE(keithw): forced escapes for '0x[hexdigit]' not implemented here
+    pattern = b'([^_a-zA-Z0-9])'
+    return re.sub(pattern, Mangle, s).decode('utf-8')
+
+
+def MangleModuleName(s):
+    # double underscores
+    return MangleName(re.sub('(_)', '__', s))
 
 
 def IsModuleCommand(command):
@@ -167,7 +177,7 @@ class CWriter(object):
                 name = re.sub(r'[^a-zA-Z0-9_]', '_', name)
                 name = os.path.splitext(name)[0]
                 self.unmangled_names[idx] = name
-                name = MangleName(name)
+                name = MangleModuleName(name)
 
                 self.module_prefix_map[idx] = name
 
@@ -178,7 +188,7 @@ class CWriter(object):
 
                 idx += 1
             elif command['type'] == 'register':
-                name = MangleName(command['as'])
+                name = MangleModuleName(command['as'])
                 if 'name' in command:
                     self.module_prefix_map[command['name']] = name
                     name_idx = self.module_name_to_idx[command['name']]
@@ -197,13 +207,13 @@ class CWriter(object):
             for line in f:
                 if 'import: ' in line:
                     line_split = line.split()
-                    import_module_name = MangleName(line_split[2][1:-1])
+                    import_module_name = MangleModuleName(line_split[2][1:-1])
                     imported_modules.add(import_module_name)
 
         if uninstantiable:
             self.out_file.write('ASSERT_TRAP(')
 
-        self.out_file.write('%s_instantiate(&%s_instance' % (self.GetModulePrefix(), self.GetModulePrefix()))
+        self.out_file.write('wasm2c_%s_instantiate(&%s_instance' % (self.GetModulePrefix(), self.GetModulePrefix()))
         for imported_module in sorted(imported_modules):
             self.out_file.write(', &%s_instance' % imported_module)
         self.out_file.write(')')
@@ -252,23 +262,21 @@ class CWriter(object):
 
     def _WriteModuleCommand(self, command):
         self.module_idx += 1
-        self.out_file.write('%s_init_module();\n' % self.GetModulePrefix())
         self._WriteModuleInitCall(command, False)
 
     def _WriteModuleInstances(self):
         idx = 0
         for command in self.commands:
             if IsModuleCommand(command):
-                self.out_file.write('%s_instance_t %s;\n' % (self.GetModulePrefix(idx), self.GetModuleInstanceName(idx)))
+                self.out_file.write('w2c_%s %s;\n' % (self.GetModulePrefix(idx), self.GetModuleInstanceName(idx)))
                 idx += 1
 
     def _WriteModuleCleanUps(self):
         for idx in range(self.module_idx):
-            self.out_file.write("%s_free(&%s_instance);\n" % (self.GetModulePrefix(idx), self.GetModulePrefix(idx)))
+            self.out_file.write("wasm2c_%s_free(&%s);\n" % (self.GetModulePrefix(idx), self.GetModuleInstanceName(idx)))
 
     def _WriteAssertUninstantiableCommand(self, command):
         self.module_idx += 1
-        self.out_file.write('%s_init_module();\n' % self.GetModulePrefix())
         self._WriteModuleInitCall(command, True)
 
     def _WriteActionCommand(self, command):
@@ -279,7 +287,19 @@ class CWriter(object):
         if len(expected) == 1:
             type_ = expected[0]['type']
             value = expected[0]['value']
-            if value == 'nan:canonical':
+            if type_ == 'v128':
+                lane_type = expected[0]['lane_type']
+                lane_count = len(expected[0]['value'])
+                # type, fmt_expected, fmt_got, f, compare, expected, found
+                self.out_file.write('ASSERT_RETURN_MULTI_T(%s, %s, %s, %s, %s, (%s), (%s));\n' %
+                                    ("v128",
+                                     " ".join("MULTI_" + ("str" if val in ('nan:canonical', 'nan:arithmetic') else lane_type) for val in value),
+                                     " ".join("MULTI_" + lane_type for _ in value),
+                                     self._Action(command),
+                                     self._SIMDCompareVector(expected[0]),
+                                     self._SIMDConstantList(expected[0]),
+                                     self._SIMDFoundList(lane_type, lane_count)))
+            elif value == 'nan:canonical':
                 assert_map = {
                     'f32': 'ASSERT_RETURN_CANONICAL_NAN_F32',
                     'f64': 'ASSERT_RETURN_CANONICAL_NAN_F64',
@@ -312,9 +332,10 @@ class CWriter(object):
             self._WriteAssertActionCommand(command)
         else:
             result_types = [result['type'] for result in expected]
-            # type, fmt, f, compare, expected, found
-            self.out_file.write('ASSERT_RETURN_MULTI_T(%s, %s, %s, %s, (%s), (%s));\n' %
+            # type, fmt_expected, fmt_got, f, compare, expected, found
+            self.out_file.write('ASSERT_RETURN_MULTI_T(%s, %s, %s, %s, %s, (%s), (%s));\n' %
                                 ("struct wasm_multi_" + MangleTypes(result_types),
+                                 " ".join("MULTI_" + ty for ty in result_types),
                                  " ".join("MULTI_" + ty for ty in result_types),
                                  self._Action(command),
                                  self._CompareList(expected),
@@ -335,16 +356,24 @@ class CWriter(object):
     def _Constant(self, const):
         type_ = const['type']
         value = const['value']
-        if type_ in ('f32', 'f64') and value in ('nan:canonical', 'nan:arithmetic'):
-            assert False
+        if type_ == 'i8':
+            return '%su' % int(value)
+        if type_ == 'i16':
+            return '%su' % int(value)
         if type_ == 'i32':
             return '%su' % int(value)
         elif type_ == 'i64':
             return '%sull' % int(value)
         elif type_ == 'f32':
+            if value in ('nan:canonical', 'nan:arithmetic'):
+                return '"(f32 %s)"' % value
             return F32ToC(int(value))
         elif type_ == 'f64':
+            if value in ('nan:canonical', 'nan:arithmetic'):
+                return '"(f64 %s)"' % value
             return F64ToC(int(value))
+        elif type_ == 'v128':
+            return 'simde_wasm_' + const['lane_type'] + 'x' + str(len(const['value'])) + '_make(' + ','.join([self._Constant({'type': const['lane_type'], 'value': x}) for x in value]) + ')'
         elif type_ == 'externref':
             if value == 'null':
                 return 'wasm_rt_externref_null_value'
@@ -375,11 +404,33 @@ class CWriter(object):
     def _CompareList(self, consts):
         return ' && '.join(self._Compare(num, const) for num, const in enumerate(consts))
 
+    def _SIMDConstantList(self, const):
+        return ', '.join(self._Constant({'type': const['lane_type'], 'value': val}) for val in const['value'])
+
+    def _SIMDFound(self, num, lane_type, lane_count):
+        return 'simde_wasm_%sx%d_extract_lane(actual, %d)' % (lane_type, lane_count, num)
+
+    def _SIMDFoundList(self, lane_type, lane_count):
+        return ', '.join(self._SIMDFound(num, lane_type, lane_count) for num in range(lane_count))
+
+    def _SIMDCompare(self, num, val, lane_type, lane_count):
+        if val == 'nan:canonical':
+            return 'is_canonical_nan_%s(%s_bits(%s))' % (lane_type, lane_type, self._SIMDFound(num, lane_type, lane_count))
+        elif val == 'nan:arithmetic':
+            return 'is_arithmetic_nan_%s(%s_bits(%s))' % (lane_type, lane_type, self._SIMDFound(num, lane_type, lane_count))
+        else:
+            return "is_equal_%s(%s, %s)" % (lane_type,
+                                            self._Constant({'type': lane_type, 'value': val}),
+                                            self._SIMDFound(num, lane_type, lane_count))
+
+    def _SIMDCompareVector(self, const):
+        return ' && '.join(self._SIMDCompare(num, val, const['lane_type'], len(const['value'])) for num, val in enumerate(const['value']))
+
     def _Action(self, command):
         action = command['action']
         type_ = action['type']
         mangled_module_name = self.GetModulePrefix(action.get('module'))
-        field = mangled_module_name + MangleName(action['field'])
+        field = "w2c_" + mangled_module_name + '_' + MangleName(action['field'])
         if type_ == 'invoke':
             args = self._ConstantList(action.get('args', []))
             if len(args) == 0:
@@ -393,13 +444,13 @@ class CWriter(object):
             raise Error('Unexpected action type: %s' % type_)
 
 
-def Compile(cc, c_filename, out_dir, *args):
+def Compile(cc, c_filename, out_dir, *cflags):
     if IS_WINDOWS:
         ext = '.obj'
     else:
         ext = '.o'
     o_filename = utils.ChangeDir(utils.ChangeExt(c_filename, ext), out_dir)
-    args = list(args)
+    args = list(cflags)
     if IS_WINDOWS:
         args += ['/nologo', '/MDd', '/c', c_filename, '/Fo' + o_filename]
     else:
@@ -441,7 +492,7 @@ def main(args):
     default_compiler = 'cc'
     if IS_WINDOWS:
         default_compiler = 'cl.exe'
-    default_compiler = os.getenv('CC', default_compiler)
+    default_compiler = os.getenv('WASM2C_CC', os.getenv('CC', default_compiler))
     parser = argparse.ArgumentParser()
     parser.add_argument('-o', '--out-dir', metavar='PATH',
                         help='output directory for files.')
@@ -452,6 +503,8 @@ def main(args):
                         help='directory to search for all executables.')
     parser.add_argument('--wasmrt-dir', metavar='PATH',
                         help='directory with wasm-rt files', default=WASM2C_DIR)
+    parser.add_argument('--simde-dir', metavar='PATH',
+                        help='directory with SIMD Everywhere files', default=SIMDE_DIR)
     parser.add_argument('--cc', metavar='PATH',
                         help='the path to the C compiler',
                         default=default_compiler)
@@ -477,9 +530,14 @@ def main(args):
     parser.add_argument('file', help='wast file.')
     parser.add_argument('--enable-exceptions', action='store_true')
     parser.add_argument('--enable-multi-memory', action='store_true')
+    parser.add_argument('--enable-memory64', action='store_true')
+    parser.add_argument('--enable-extended-const', action='store_true')
     parser.add_argument('--disable-bulk-memory', action='store_true')
     parser.add_argument('--disable-reference-types', action='store_true')
     parser.add_argument('--debug-names', action='store_true')
+    parser.add_argument('--num-outputs', metavar='COUNT',
+                        help='number of output c files for wasm2c', dest='num_outputs',
+                        default=1, type=int, action='store')
     options = parser.parse_args(args)
 
     with utils.TempDirectory(options.out_dir, 'run-spec-wasm2c-') as out_dir:
@@ -491,6 +549,8 @@ def main(args):
         wast2json.AppendOptionalArgs({
             '-v': options.verbose,
             '--enable-exceptions': options.enable_exceptions,
+            '--enable-memory64': options.enable_memory64,
+            '--enable-extended-const': options.enable_extended_const,
             '--enable-multi-memory': options.enable_multi_memory,
             '--disable-bulk-memory': options.disable_bulk_memory,
             '--disable-reference-types': options.disable_reference_types,
@@ -506,6 +566,8 @@ def main(args):
         wasm2c.verbose = options.print_cmd
         wasm2c.AppendOptionalArgs({
             '--enable-exceptions': options.enable_exceptions,
+            '--enable-memory64': options.enable_memory64,
+            '--enable-extended-const': options.enable_extended_const,
             '--enable-multi-memory': options.enable_multi_memory})
 
         options.cflags += shlex.split(os.environ.get('WASM2C_CFLAGS', ''))
@@ -525,15 +587,28 @@ def main(args):
         cwriter = CWriter(spec_json, prefix, output, out_dir)
 
         o_filenames = []
-        includes = '-I%s' % options.wasmrt_dir
+        cflags = ['-I%s' % options.wasmrt_dir, '-I%s' % options.simde_dir]
+        if options.enable_memory64:
+            if IS_WINDOWS:
+                sys.stderr.write('skipping: wasm2c+memory64 is not yet supported under msvc\n')
+                return SKIPPED
+            cflags.append('-DSUPPORT_MEMORY64=1')
 
         for i, wasm_filename in enumerate(cwriter.GetModuleFilenames()):
             wasm_filename = os.path.join(out_dir, wasm_filename)
-            c_filename = utils.ChangeExt(wasm_filename, '.c')
-            args = ['-n', cwriter.GetModulePrefixUnmangled(i)]
-            wasm2c.RunWithArgs(wasm_filename, '-o', c_filename, *args)
+            c_filename_input = utils.ChangeExt(wasm_filename, '.c')
+            c_filenames = []
+            if options.num_outputs > 1:
+                base = os.path.splitext(c_filename_input)[0]
+                for j in range(options.num_outputs):
+                    c_filenames.append(base + '_' + str(j) + '.c')
+            else:
+                c_filenames.append(utils.ChangeExt(wasm_filename, '.c'))
+            args = ['-n', cwriter.GetModulePrefixUnmangled(i), '--num-outputs', str(options.num_outputs)]
+            wasm2c.RunWithArgs(wasm_filename, '-o', c_filename_input, *args)
             if options.compile:
-                o_filenames.append(Compile(cc, c_filename, out_dir, includes))
+                for j, c_filename in enumerate(c_filenames):
+                    o_filenames.append(Compile(cc, c_filename, out_dir, *cflags))
 
         cwriter.Write()
         main_filename = utils.ChangeExt(json_file_path, '-main.c')
@@ -543,10 +618,14 @@ def main(args):
         if options.compile:
             # Compile wasm-rt-impl.
             wasm_rt_impl_c = os.path.join(options.wasmrt_dir, 'wasm-rt-impl.c')
-            o_filenames.append(Compile(cc, wasm_rt_impl_c, out_dir, includes))
+            o_filenames.append(Compile(cc, wasm_rt_impl_c, out_dir, *cflags))
+
+            # Compile wasm-rt-exceptions.
+            wasm_rt_exceptions_c = os.path.join(options.wasmrt_dir, 'wasm-rt-exceptions-impl.c')
+            o_filenames.append(Compile(cc, wasm_rt_exceptions_c, out_dir, *cflags))
 
             # Compile and link -main test run entry point
-            o_filenames.append(Compile(cc, main_filename, out_dir, includes))
+            o_filenames.append(Compile(cc, main_filename, out_dir, *cflags))
             if IS_WINDOWS:
                 exe_ext = '.exe'
                 libs = []
