@@ -384,6 +384,7 @@ class TestInfo(object):
         self.slow = False
         self.skip = False
         self.is_roundtrip = False
+        self.is_wasm2c = False
 
     def CreateRoundtripInfo(self, fold_exprs):
         if self.tool not in ROUNDTRIP_TOOLS:
@@ -453,6 +454,7 @@ class TestInfo(object):
         if tool not in TOOLS:
             raise Error('Unknown tool: %s' % tool)
         self.tool = tool
+        self.is_wasm2c = self.tool == 'run-spec-wasm2c'
         for tool_key, tool_value in TOOLS[tool]:
             self.ParseDirective(tool_key, tool_value)
 
@@ -666,7 +668,7 @@ class Status(object):
         assert(self.isatty)
         total_duration = time.time() - self.start_time
         name = info.GetName() if info else ''
-        if (self.total - self.skipped):
+        if self.total - self.skipped:
             percent = 100 * (self.passed + self.failed) / (self.total - self.skipped)
         else:
             percent = 100
@@ -784,6 +786,11 @@ def HandleTestResult(status, info, result, rebase=False):
                 status.Passed(info, result.duration)
         else:
             if result.Failed():
+                if result.GetLastFailure().returncode == 3:
+                    # run-spec-wasm2c.py returns 3 to signal that the test
+                    # should be skipped.
+                    status.Skipped(info)
+                    return
                 # This test has already failed, but diff it anyway.
                 last_failure = result.GetLastFailure()
                 msg = 'expected error code %d, got %d.' % (
